/* SR3C, a symbol ranking data compressor.
 *
 * A simple test compression/decompression with fragmenting/merging
 * packets and exchange of sending/receiving parties.  Reads test data
 * input from stdin.
 *
 * Author: Kenneth Oksanen <cessu@iki.fi>, 2008.
 * Copyright (C) Helsinki University of Technology.
 *
 * This file is distributed under the following license:

The MIT License
Copyright (c) 2008 Helsinki University of Technology
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

 */


/* A test which takes input from stdin, chops it into small pieces,
   compresses (and possibly duplicates) and randomly switches
   compressing and uncompressing sides.  Checks whether the data was
   received correct and whether the significant parts of sending and
   receiving contexts were intact. */


/* Some controls on what is being tested. */
#define DUPLICATE  1
#define FRAGMENT   1


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* Include the source code so that we can check various parts if the
   internal state of it.
 */
#include "sr3c.c"


unsigned char msg_queue[10000], *msg_p;

int send_f(const unsigned char *bytes, size_t n_bytes, int flushed, void *ctx)
{
  while (n_bytes--)
    *msg_p++ = *bytes++;
  /* fprintf(stderr, "send_f flushed = %d\n", flushed); */
  return 0;
}

unsigned char plain_queue[10000], *plain_p;

int recv_f(const unsigned char *bytes, size_t n_bytes, int flushed, void *ctx)
{
  while (n_bytes--)
    *plain_p++ = *bytes++;
  /* fprintf(stderr, "recv_f flushed = %d\n", flushed); */
  return 0;
}

int main(void)
{
#define FAILSTEP  201450000
  int i, j, step;
  unsigned char buf[1024];
  sr3c_context_t *compress, *uncompress, *tmp;
  sr3c_output_f_t tmp_f;
  size_t n_bytes, res_n_bytes;

  srandom(0);

  compress = sr3c_alloc(send_f, NULL);
  uncompress = sr3c_alloc(recv_f, NULL);

  while (!feof(stdin)) {
    /* Chop the input randomly. */
    n_bytes = random() % 4;
    n_bytes = fread(buf, 1, n_bytes, stdin);

    /* memset used only to distinguish buffer overruns in debugging. */
    memset(msg_queue, 0x55, sizeof(msg_queue));

    if (step > FAILSTEP) {
      fprintf(stderr, "[");
      for (i = 0; i < n_bytes; i++)
	fprintf(stderr, "%02X", buf[i]);
      fprintf(stderr, "] -> [");
    }

    msg_p = &msg_queue[0];
    sr3c_compress(buf, n_bytes, compress);
    sr3c_flush(compress);
#ifdef DUPLICATE
    sr3c_compress(buf, n_bytes, compress);
    sr3c_flush(compress);
    res_n_bytes = 2 * n_bytes;
#else
    res_n_bytes = n_bytes;
#endif
    if (step > FAILSTEP) {
      for (i = 0; i < msg_p - msg_queue; i++)
	fprintf(stderr, "%02X", msg_queue[i]);
      fprintf(stderr, "]\n");
    }

    plain_p = &plain_queue[0];
#ifdef FRAGMENT
    {
      int n_left = msg_p - msg_queue, n;
      unsigned char *p = msg_queue;

      while (n_left) {
	n = random() % 6;
	if (n > n_left)
	  n = n_left;
	if (step > FAILSTEP)
	  fprintf(stderr, "n = %d, n_left = %d, status = %d, %02X %02X %02X %02X %02X\n",
		  n, n_left, uncompress->status,
		  p[0], p[1], p[2], p[3], p[4]);
	sr3c_uncompress(p, n, uncompress);
	p += n;
	n_left -= n;
      }
      if (step > FAILSTEP)
	fprintf(stderr, "status = %d\n", uncompress->status);
    }
#else
    sr3c_uncompress(msg_queue, msg_p - msg_queue, uncompress);
#endif

    if (1 || step > FAILSTEP) {
      if (compress->hash != uncompress->hash)
	fprintf(stderr, "hash = 0x%08X vs 0x%08X\n",
		compress->hash, uncompress->hash);
      if (compress->prev_ch != uncompress->prev_ch)
	fprintf(stderr, "prev_ch = 0x%2X vs 0x%02X\n",
		compress->prev_ch, uncompress->prev_ch);
      if (compress->x1 != uncompress->x1)
	fprintf(stderr, "x1 = 0x%08X vs 0x%08X\n",
		compress->x1, uncompress->x1);
      if (compress->x2 != uncompress->x2)
	fprintf(stderr, "x2 = 0x%08X vs 0x%08X\n",
		compress->x2, uncompress->x2);
      for (i = 0; i < SR_HMASK + 1; i++)
	if (compress->rank[i] != uncompress->rank[i])
	  fprintf(stderr, "rank[%d] = 0x%08X vs 0x%08X\n", i,
		  compress->rank[i], uncompress->rank[i]);
      for (i = 0; i < SM_HUFF_LO_N; i++)
	if (compress->sm_huff_lo[i] != uncompress->sm_huff_lo[i])
	  fprintf(stderr, "huff_lo[%d] = 0x%08X vs 0x%08X\n", i,
		  compress->sm_huff_lo[i], uncompress->sm_huff_lo[i]);
      for (i = 0; i < SM_HUFF_HI_N1; i++)
	for (j = 0; j < SM_HUFF_HI_N2; j++)
	  if (compress->sm_huff_hi[i][j] != uncompress->sm_huff_hi[i][j])
	    fprintf(stderr, "huff_hi[%d][%d] = 0x%08X vs 0x%08X\n", i, j,
		    compress->sm_huff_hi[i][j], uncompress->sm_huff_hi[i][j]);
      for (i = 0; i < SM_BYTES_N; i++)
	if (compress->sm_bytes[i] != uncompress->sm_bytes[i])
	  fprintf(stderr, "bytes[%d] = 0x%08X vs 0x%08X\n", i,
		  compress->sm_bytes[i], uncompress->sm_bytes[i]);
    }
    
    if (step > FAILSTEP)
      fprintf(stderr, "%d -> %d -> %d   %s %s\n", n_bytes, msg_p - msg_queue,
	      plain_p - plain_queue,
	      plain_p - plain_queue == res_n_bytes ? "OK" : "FAIL",
	      memcmp(buf, plain_queue, n_bytes) == 0 ? "OK" : "FAIL");

    if (plain_p - plain_queue != res_n_bytes
	|| memcmp(buf, plain_queue, n_bytes) != 0) {
      fprintf(stderr, "FAIL at step = %d\n", step);
      return 1;
    }

    if (random() % 2) {
      if (step > FAILSTEP)
	fprintf(stderr, "SWAP\n");
      /* Swap the compressing and uncompressing context, as if to
	 simulate a bidirectional compressed message interchange. */
      tmp = compress;
      compress = uncompress;
      uncompress = tmp;
      /* ... but retain the output functions as they were. */
      tmp_f = compress->output_f;
      compress->output_f = uncompress->output_f;
      uncompress->output_f = tmp_f;
    }

    step++;
  }

  return 0;
}
